extern "C" {
#include <eadk.h>
#include "k_defs.h"
}
#include <cstring>
#include <cstdlib>
#include <string>
#include <cassert>

const int heapSize=128*1024; // heap size for the app called
const size_t tar_flash_addresses[]={0x90200000u,0x90400000};
const size_t tar_flash_maxsizes[]={0x60000,0x400000};
#define ENTRYPOINT 0x90400200

extern const char eadk_app_name[] __attribute__((section(".rodata.eadk_app_name"))) = "KhiCAS";
extern const uint32_t eadk_api_level __attribute__((section(".rodata.eadk_api_level"))) = 0;

// missing in EADK
#include "svcall.h"
#define SVC_ATTRIBUTES __attribute__((noinline, externally_visible))

class ShiftAlphaStatus {
public:
  enum class ShiftStatus : bool { Inactive = false, Active = true };
  enum class AlphaStatus : uint8_t {
    Inactive = 0,
    Active,
    Locked,
    NumberOfStatuses
  };

  constexpr ShiftAlphaStatus(ShiftStatus shift, AlphaStatus alpha)
      : m_value{.fields = {.shift = shift, .alpha = alpha}} {}

  constexpr ShiftAlphaStatus()
      : ShiftAlphaStatus(ShiftStatus::Inactive, AlphaStatus::Inactive) {}

  operator uint8_t() const { return m_value.raw; }

  bool operator==(const ShiftAlphaStatus& other) {
    return m_value.raw == other.m_value.raw;
  }
  bool operator!=(const ShiftAlphaStatus& other) { return !(*this == other); }

  bool shiftIsActive() const {
    return m_value.fields.shift != ShiftStatus::Inactive;
  }
  bool alphaIsActive() const {
    return m_value.fields.alpha != AlphaStatus::Inactive;
  }
  bool alphaIsLocked() const {
    return m_value.fields.alpha == AlphaStatus::Locked;
  }

  void removeShift() { m_value.fields.shift = ShiftStatus::Inactive; }
  void removeAlpha() { m_value.fields.alpha = AlphaStatus::Inactive; }

  void toggleShift() {
    m_value.fields.shift =
        static_cast<ShiftStatus>(!static_cast<bool>(m_value.fields.shift));
  }
  void cycleAlpha() {
    m_value.fields.alpha = static_cast<AlphaStatus>(
        (static_cast<uint8_t>(m_value.fields.alpha) + 1) %
        static_cast<uint8_t>(AlphaStatus::NumberOfStatuses));
    if (m_value.fields.alpha == AlphaStatus::Inactive) {
      removeShift();
    }
  }

  // private:
  /* This is implemented as an uint8_t so that only 1 register has to be passed
   * through SVCalls. */
  union {
    uint8_t raw;
    struct {
      ShiftStatus shift : 1;
      AlphaStatus alpha : 2;
    } fields;
  } m_value;
};
// end svcall related

#define API_VERSION 2
#define EXTAPP_RAM_FILE_SYSTEM 0
#define EXTAPP_FLASH_FILE_SYSTEM 1
#define EXTAPP_BOTH_FILE_SYSTEM 2
typedef uint32_t (*entrypoint)(const uint32_t, const void *, void *, const uint32_t);

static char print_hex(int val){
  val &= 0xf;
  if (val>=0 && val<=9)
    return '0'+val;
  return 'A'+(val-10);
}

static void print_hex(unsigned val,char * ptr){
  for (int i=0;i<8;++i){
    ptr[i]=print_hex(val>>(28-4*i));
  }
}
  
extern "C" {
  ShiftAlphaStatus SVC_ATTRIBUTES getShiftAlphaStatus() {
    SVC_RETURNING_R0(SVC_EVENTS_SHIFT_ALPHA_STATUS, ShiftAlphaStatus)}
  
  void SVC_ATTRIBUTES setShiftAlphaStatus(ShiftAlphaStatus s){
    SVC_RETURNING_VOID(SVC_EVENTS_SET_SHIFT_ALPHA_STATUS)}

  void extapp_lockAlpha() {
    ShiftAlphaStatus s(ShiftAlphaStatus::ShiftStatus::Inactive, ShiftAlphaStatus::AlphaStatus::Locked);
    setShiftAlphaStatus(s);
  }

  void extapp_resetKeyboard() {
    ShiftAlphaStatus s(ShiftAlphaStatus::ShiftStatus::Inactive, ShiftAlphaStatus::AlphaStatus::Inactive);
    setShiftAlphaStatus(s);
  }

  inline uint32_t reverse32(uint32_t value) {
    return (((value & 0x000000FF) << 24) |
            ((value & 0x0000FF00) <<  8) |
            ((value & 0x00FF0000) >>  8) |
            ((value & 0xFF000000) >> 24));
  }

  // inspire by storage.c by Yaya.Cout
  // https://framagit.org/Yaya.Cout/numworks-extapp-storage/-/blob/master/src/storage.c?ref_type=heads
  const int extapp_calculatorModel() {
    // check userland header    
    size_t * storage1 = (size_t *)0x90010000 ;
    size_t * storage2 = (size_t *)0x90020000 ;
    if (*storage1==0xdec0edfe)
      return 1;
    if (*storage2==0xdec0edfe)
      return 2;
    return 0;
  }

  const size_t * extapp_userlandAddress() {
    size_t * storage1 = (size_t *)0x90010000 ;
    size_t * storage2 = (size_t *)0x90020000 ;
    if (*storage1==0xdec0edfe)
      return storage1;
    if (*storage2==0xdec0edfe)
      return storage2;
    return 0;
  }

  unsigned char * storage_address() {
    const size_t * ptr= extapp_userlandAddress();
    if (!ptr)
      return 0;
    return (unsigned char *) (*(ptr+3));
  }
  
  const size_t storage_size() {
    const size_t * ptr= extapp_userlandAddress();
    if (!ptr)
      return 0;
    return *(ptr+4);
  }
  
  bool extapp_isValid(const uint32_t * address) {
    return *address == 0xee0bddba;
  }
  
  bool erase_record(const char * name){
    unsigned char * ptr=(unsigned char *)storage_address();
    if (!ptr)
      return false;
    for (ptr+=4;;){ // skip header
      size_t L=ptr[1]*256+ptr[0]; 
      if (L==0) return false;
      if (strcmp((const char *)ptr+2,name)==0){
        unsigned char * newptr=ptr;
        int S=0,erased=L;
        ptr+=L;
        for (;;){
          L=ptr[1]*256+ptr[0];
          if (L==0){
            for (int i=0;i<erased;++i)
              newptr[S+i]=0;
            break;
          }
          // if (ptr>newptr+S)
          memmove(newptr+S,ptr,L);
          S+=L;
          ptr+=L; 
        }
        return true;
      }
      ptr+=L;
    }
  }
    
  bool write_record(const char * name,const void * content,int len,bool write0){
    unsigned char * ptr=(unsigned char *)storage_address();
    if (!ptr)
      return false;
    unsigned char * debut=ptr;
    int taille=storage_size();
    // first erase record if it exists
    erase_record(name);
    for (ptr+=4;;){ // skip header
      size_t L=ptr[1]*256+ptr[0]; 
      if (L!=0){
        ptr+=L;
        continue;
      }
      int l=strlen(name)+1;
      L=l+len+2; // total record length
      if (write0) ++L;
      if (ptr-debut+L>=taille-2) 
        return false; // overflow
      ptr[1]=L/256;
      ptr[0]=L%256;
      ptr+=2;
      strncpy((char*)ptr,name,l);
      ptr+=l;
      if (write0){
        *ptr=0;
        ++ptr;
      }
      memcpy(ptr,content,len);
      ptr+=len;
      ptr[0]=0;
      ptr[1]=0;
      return true;
    }
  }
    
  bool extapp_inexammode() {
    return false;
  }
  
  void sync_screen() {
  }
  

  uint64_t extapp_millis() {
    return eadk_timing_millis();
  }

  void extapp_msleep(uint32_t ms) {
    eadk_timing_msleep(ms);
  }

  uint64_t extapp_scanKeyboard() {
    return eadk_keyboard_scan();
  }

  void extapp_pushRect(int16_t x, int16_t y, uint16_t w, uint16_t h, const uint16_t * pixels) {
    eadk_display_push_rect((eadk_rect_t){(uint16_t)x, (uint16_t)(y),w,h},pixels);
  }
  
  void extapp_pushRectUniform(int16_t x, int16_t y, uint16_t w, uint16_t h, uint16_t color) {
#if 0
    if (x<0) {
      w += x;
      x=0;
    }
    if (y<0) {
      h += y;
      y=0;
    }
    if (h+y>=EADK_SCREEN_HEIGHT)
      h=EADK_SCREEN_HEIGHT-y;
    if (x+w>=EADK_SCREEN_WIDTH)
      w=EADK_SCREEN_WIDTH-x;
    if (h<=0 || w<=0)
      return;
#else
    if (x<0 || x>=EADK_SCREEN_WIDTH || y<0 || y>=EADK_SCREEN_HEIGHT)
      return;
#endif
    eadk_display_push_rect_uniform((eadk_rect_t){(uint16_t)x, (uint16_t)(y),w,h},(eadk_color_t)color);
  }

  void extapp_pullRect(int16_t x, int16_t y, uint16_t w, uint16_t h, uint16_t * pixels) {
    eadk_display_pull_rect((eadk_rect_t){(uint16_t)x,(uint16_t)y,1,1}, pixels);
  }

  int16_t extapp_drawTextLarge(const char *text, int16_t x, int16_t y, uint16_t fg, uint16_t bg, bool fake) {
    if (!fake) {
      eadk_display_draw_string(text, (eadk_point_t){(uint16_t)x,(uint16_t)(y)}, true, (eadk_color_t)fg, (eadk_color_t)bg);
    }
    return x+strlen(text)*10;
  }

  int16_t extapp_drawTextSmall(const char *text, int16_t x, int16_t y, uint16_t fg, uint16_t bg, bool fake) {
    if (!fake) {
      eadk_display_draw_string(text, (eadk_point_t){(uint16_t)x,(uint16_t)(y)}, false, (eadk_color_t)fg, (eadk_color_t)bg);
    }
    return x+strlen(text)*7;
  }

  bool extapp_waitForVBlank() {
    return eadk_display_wait_for_vblank();
  }

  size_t SVC_ATTRIBUTES copyText(uint8_t eventId, char* buffer,
                                 size_t bufferSize) {
    SVC_RETURNING_R0(SVC_EVENTS_COPY_TEXT, size_t)}
  
  void extapp_clipboardStore(const char *text) {
  }

  const char * extapp_clipboardText() {
    return "";
  }
  
  bool match(const char * filename,const char * extension){
    if (!extension)
      return true;
    int el=strlen(extension);
    int fl=strlen(filename);
    if (el>=fl) return false;
    int i=fl-el,j=0;
    for (;j<el;++i,++j){
      if (filename[i]!=extension[j])
        return false;
    }
    return true;
  }
  
  bool extapp_fileErase(const char * filename, int storage) {
    if (storage==EXTAPP_RAM_FILE_SYSTEM) {
      return erase_record(filename);
    } 
    return false;
  }

  int filesize(const char * buffer,size_t header_offset) {
    // offset: 124
    const char * szView = buffer+ header_offset+124;
    int res=0;
    for (int i = 0; i < 11; i++) {
      char tmp=szView[i];
      if (tmp<'0' || tmp>'9') return -1; // invalid file size
      res *= 8;
      res += (tmp-'0');
    }
    return res;
  }

  int tar_filesize(int s){
    int h=(s/512+1)*512;
    if (s%512)
      h +=512;
    return h;
  }
  
  int tar_mode(const char * buffer,size_t header_offset) {
    // offset: 100
    const char * szView = buffer+ header_offset+100;
    int res=0;
    for (int i = 0; i < 7; i++) {
      char tmp=szView[i];
      if (tmp==' ')
        return res;
      if (tmp<'0' || tmp>'9') 
        return -1; // invalid file size
      res *= 10;
      res += (tmp-'0');
    }
    return res;
  }

  const char * flash_fileread(const char * buffer,size_t byteLength,const char * filename){
    if (!buffer)
      return 0;
    size_t offset=0;       
    // skip user apps
    while ( ( ((unsigned char) buffer[offset])==0xba
              || ((unsigned char) buffer[offset+1])==0xbe
              ) && offset<byteLength){
      offset+=0x10000;
    }
    while (byteLength==0 || offset<byteLength-512){
      // file_name
      const char * file_name = buffer+offset;
      if (*file_name == 0) 
        return 0;
      int s=filesize(buffer,offset);
      if (s<0)
        return 0;
      int mode=tar_mode(buffer,offset);
      size_t oldoffset=offset;
      offset += tar_filesize(s);
      if (mode<0)
        continue;
      mode = (mode/100)&4; 
      if (mode==0)
        continue;
      if (strcmp(file_name,filename)==0)
        return buffer+oldoffset+512;
    }
    return 0;
  }

  int extapp_getKey(int allow_suspend, bool *alphawasactive) ;
  
  int extapp_fileListWithExtension(const char ** filenames, int maxrecords, const char * extension, int storage) {
    int j=0;
    if (storage==EXTAPP_RAM_FILE_SYSTEM || storage==EXTAPP_BOTH_FILE_SYSTEM) {
      unsigned char * ptr=(unsigned char *)storage_address();
      if (ptr){
        for (ptr+=4;j<maxrecords;){
          size_t L=ptr[1]*256+ptr[0]; 
          if (L==0) break;
          if (match((const char *)ptr+2,extension)){
            filenames[j]=(const char *)ptr+2;
            ++j;
          }
          ptr+=L;
        }
      }
    }
    if (storage==EXTAPP_FLASH_FILE_SYSTEM  || storage==EXTAPP_BOTH_FILE_SYSTEM) {
      for (int i=0;i<sizeof(tar_flash_addresses)/sizeof(size_t);i++){
        size_t offset=0;
        const char * buffer = (const char *) tar_flash_addresses[i];
        int byteLength=tar_flash_maxsizes[i];
        // skip user apps
        while ( ( ((unsigned char) buffer[offset])==0xba
                  || ((unsigned char) buffer[offset+1])==0xbe
                  ) && offset<byteLength){
          offset+=0x10000;
        }
        while (j<maxrecords && offset<byteLength-512){
          const char * file_name = buffer+offset;
          // file_name
          if (*file_name==0 || strnlen(file_name,33)==33) 
            break;
          int s=filesize(buffer,offset);
          if (s<0)
            return 0;
          int mode=tar_mode(buffer,offset);
          offset += tar_filesize(s);
#if 0
          extapp_drawTextSmall("File            ",0,120,65535,0,false);
          extapp_drawTextSmall(file_name,60,120,65535,0,false);         
          extapp_drawTextSmall("Mode            ",0,140,65535,0,false);
          char tab3[9]={0};
          sprintf(tab3,"%d",mode);
          extapp_drawTextSmall(tab3,60,140,65535,0,false);
          extapp_getKey(false,0);
#endif
          if (mode<0)
            continue;
          mode = (mode/100) & 4; 
          if (mode==0)
            continue;
          if (match(file_name,extension)){
            filenames[j]=file_name;
            ++j;
          }
        }
      }
    }
    return j;
  }

  const char * extapp_fileRead(const char * filename, size_t *len, int storage) {
    if (storage==EXTAPP_RAM_FILE_SYSTEM || storage==EXTAPP_BOTH_FILE_SYSTEM) {
      unsigned char * ptr=(unsigned char *)storage_address();
      if (!ptr)
        return 0;
      for (ptr+=4;;){ // skip header
        size_t L=ptr[1]*256+ptr[0]; 
        if (L==0) break;
        if (strcmp((const char *)ptr+2,filename)==0){
          // skip filename and initial 0 or 1 from script
          if (len) *len=L-(strlen(filename)+1 +1); 
          ptr += strlen(filename)+3 +1;
          return (const char *) ptr;
        }
        ptr += L;
      }
    }
    if (storage==EXTAPP_FLASH_FILE_SYSTEM || storage==EXTAPP_BOTH_FILE_SYSTEM) {
      for (int i=0;i<sizeof(tar_flash_addresses)/sizeof(size_t);++i){
        const char * ptr=flash_fileread((char *)tar_flash_addresses[i],tar_flash_maxsizes[i],filename);
        if (ptr) return ptr;
      }
    } 
    return NULL;
  }

  bool extapp_fileExists(const char * filename, int storage) {
    const char * ptr=extapp_fileRead(filename,0,EXTAPP_BOTH_FILE_SYSTEM);
    return ptr;
  }
  
  bool extapp_fileWrite(const char * filename, const char * content, size_t len, int storage) {
    if (storage == EXTAPP_RAM_FILE_SYSTEM) {
      return write_record(filename,content,len,false /* write 0 at begin*/);
    }
    // not handled by the launcher, there is no buffer for that (64K required)
    return false;
  }
  
  int extapp_restorebackup(int mode){
    return 0;
  }

  // info from ion/src/device/include/n0110/config/external_flash.h
  // the files for n0115 and n0120 are identical
  constexpr int NumberOf4KSectors = 8;
  constexpr int NumberOf32KSectors = 1;
  constexpr int NumberOf64KSectors = 128 - 1;
  constexpr int NumberOfSectors =
    NumberOf4KSectors + NumberOf32KSectors + NumberOf64KSectors;
  
  constexpr uint8_t NumberOfAddressBitsIn64KbyteBlock = 16;
  constexpr uint8_t NumberOfAddressBitsIn32KbyteBlock = 15;
  constexpr uint8_t NumberOfAddressBitsIn4KbyteBlock = 12;

  // borrowed from ion/src/device/shared/drivers/flash_information.cpp
  int TotalNumberOfSectors() { return NumberOfSectors; }
  
  int SectorAtAddress(uint32_t address) {
    /* WARNING: this code assumes that the flash sectors are of increasing size:
     * first all 4K sectors, then all 32K sectors, and finally all 64K sectors. */
    address -= 0x90000000; // Board::Config::ExternalFlashOrigin;
    int i = address >> NumberOfAddressBitsIn64KbyteBlock;
    if (i > NumberOf64KSectors) {
      return -1;
    }
    if (i >= 1) {
      return NumberOf4KSectors +
        NumberOf32KSectors + i - 1;
    }
    i = address >> NumberOfAddressBitsIn32KbyteBlock;
    if (i >= 1) {
      assert(i >= 0 && i <= NumberOf32KSectors);
      i = NumberOf4KSectors + i - 1;
      return i;
    }
    i = address >> NumberOfAddressBitsIn4KbyteBlock;
    assert(i <= NumberOf4KSectors);
    return i;
  }

  // inside dfu handleInterrupts is set to false by Epsilon
  // inside writing exam mode, it is set to true
  bool SVC_ATTRIBUTES EraseSectorWithInterruptions(int i, bool handleInterrupts) {
  SVC_RETURNING_R0(SVC_FLASH_ERASE_SECTOR_WITH_INTERRUPTIONS, bool)}

  bool SVC_ATTRIBUTES WriteMemoryWithInterruptions(uint8_t* destination,
                                                 const uint8_t* source,
                                                 size_t length,
                                                 bool handleInterrupts) {
  SVC_RETURNING_R0(SVC_FLASH_WRITE_MEMORY_WITH_INTERRUPTIONS, bool)}
  
  void extapp_erasesector(const char *ptr){
    EraseSectorWithInterruptions(SectorAtAddress((uint32_t) ptr),true);
  }
  
  bool extapp_writememory(unsigned char * dest,const unsigned char * data,size_t length){
    return WriteMemoryWithInterruptions(dest,data,length,true);
  }

  bool waitforvblank() {
    return eadk_display_wait_for_vblank();
  }


  static const int translated_keys[]=
    {
      // non shifted
      KEY_CTRL_LEFT,KEY_CTRL_UP,KEY_CTRL_DOWN,KEY_CTRL_RIGHT,KEY_CTRL_OK,KEY_CTRL_EXIT,
      KEY_CTRL_MENU,KEY_PRGM_ACON,KEY_PRGM_ACON,9,10,11,
      KEY_CTRL_SHIFT,KEY_CTRL_ALPHA,KEY_CTRL_XTT,KEY_CTRL_VARS,KEY_CTRL_CATALOG,KEY_CTRL_DEL,
      KEY_CHAR_EXPN,KEY_CHAR_LN,KEY_CHAR_LOG,KEY_CHAR_IMGNRY,',',KEY_CHAR_POW,
      KEY_CHAR_SIN,KEY_CHAR_COS,KEY_CHAR_TAN,KEY_CHAR_PI,KEY_CHAR_ROOT,KEY_CHAR_SQUARE,
      '7','8','9','(',')',-1,
      '4','5','6','*','/',-1,
      '1','2','3','+','-',-1,
      '0','.',KEY_CHAR_EXPN10,KEY_CHAR_ANS,KEY_CTRL_EXE,-1,
      // shifted
      KEY_SHIFT_LEFT,KEY_CTRL_PAGEUP,KEY_CTRL_PAGEDOWN,KEY_SHIFT_RIGHT,KEY_CTRL_OK,KEY_CTRL_EXIT,
      KEY_CTRL_MENU,KEY_PRGM_ACON,KEY_PRGM_ACON,9,10,11,
      KEY_CTRL_SHIFT,KEY_CTRL_ALPHA,KEY_CTRL_CUT,KEY_CTRL_CLIP,KEY_CTRL_PASTE,KEY_CTRL_AC,
      KEY_CHAR_LBRCKT,KEY_CHAR_RBRCKT,KEY_CHAR_LBRACE,KEY_CHAR_RBRACE,'_',KEY_CHAR_STORE,
      KEY_CHAR_ASIN,KEY_CHAR_ACOS,KEY_CHAR_ATAN,'=','<','>',
      KEY_CTRL_F7,KEY_CTRL_F8,KEY_CTRL_F9,KEY_CTRL_F13,KEY_CTRL_F14,-1,
      KEY_CTRL_F4,KEY_CTRL_F5,KEY_CTRL_F6,KEY_CTRL_INS,'%',-1,
      KEY_CTRL_F1,KEY_CTRL_F2,KEY_CTRL_F3,'\t','\\',-1,
      KEY_CTRL_F10,KEY_CTRL_F11,KEY_CTRL_F12,KEY_SHIFT_ANS,KEY_CTRL_MENU,-1,
      // alpha
      KEY_CTRL_LEFT,KEY_CTRL_UP,KEY_CTRL_DOWN,KEY_CTRL_RIGHT,KEY_CTRL_OK,KEY_CTRL_EXIT,
      KEY_CTRL_MENU,KEY_PRGM_ACON,KEY_PRGM_ACON,9,10,11,
      KEY_CTRL_SHIFT,KEY_CTRL_ALPHA,':',';','"',KEY_CTRL_DEL,
      'a','b','c','d','e','f',
      'g','h','i','j','k','l',
      'm','n','o','p','q',-1,
      'r','s','t','u','v',-1,
      'w','x','y','z',' ',-1,
      '?','!',KEY_CHAR_EXPN10,KEY_CHAR_ANS,KEY_CTRL_EXE,-1,
      // alpha shifted
      KEY_SHIFT_LEFT,KEY_CTRL_PAGEUP,KEY_CTRL_PAGEDOWN,KEY_SHIFT_RIGHT,KEY_CTRL_OK,KEY_CTRL_EXIT,
      KEY_CTRL_MENU,KEY_PRGM_ACON,KEY_PRGM_ACON,9,10,11,
      KEY_CTRL_SHIFT,KEY_CTRL_ALPHA,':',';','\'','%',
      'A','B','C','D','E','F',
      'G','H','I','J','K','L',
      'M','N','O','P','Q',-1,
      'R','S','T','U','V',-1,
      'W','X','Y','Z',' ',-1,
      '?','!',KEY_CHAR_EXPN10,KEY_CHAR_ANS,KEY_CTRL_EXE,-1,
    };

  int extapp_kbdstatus(){
    return getShiftAlphaStatus();
  }
  
  int getkey(int ck_keydown,int32_t timeout=1000) {
    if (ck_keydown & 0x80000000){
      // iskeydown call, will work for cursor OK and back keys
      int k=ck_keydown & 0xff;
      if (eadk_keyboard_key_down(eadk_keyboard_scan(), (eadk_key_t) k)) 
        return 1;
      else
        return 0;
    }
    bool shifted=getShiftAlphaStatus() & 1;
    eadk_event_t e = eadk_event_get(&timeout);
    if (e == eadk_event_shift_left && eadk_keyboard_key_down(eadk_keyboard_scan(), eadk_key_shift)) 
      return KEY_SELECT_LEFT;
    if (e == eadk_event_shift_right && eadk_keyboard_key_down(eadk_keyboard_scan(), eadk_key_shift)) 
      return KEY_SELECT_RIGHT;
    if (e == eadk_event_shift_up && eadk_keyboard_key_down(eadk_keyboard_scan(), eadk_key_shift)) 
      return KEY_SELECT_UP;
    if (e == eadk_event_shift_down && eadk_keyboard_key_down(eadk_keyboard_scan(), eadk_key_shift)) 
      return KEY_SELECT_DOWN;
    // Workaround to make shift-number work
    if (e >= eadk_event_seven && e <= eadk_event_exe) {
      if (
          shifted
          //eadk_keyboard_key_down(eadk_keyboard_scan(), eadk_key_shift)
          ) {
        e += 54;
      }
    }
    return translated_keys[e];
  }

  void GetKey(int * key) {
    *key = getkey(true);
  }

  int extapp_getKey(int ck_keydown, bool *alphawasactive) {
    if (alphawasactive){
      *alphawasactive=getShiftAlphaStatus().alphaIsActive();
    }
    return getkey(ck_keydown,900000);
  }

  bool extapp_isalphaactive(){
    //return false;
    return getShiftAlphaStatus().alphaIsActive();
  }
  
  extern "C" void (* const apiPointers[])(void) = {
    (void (*)(void)) extapp_millis,
    (void (*)(void)) extapp_msleep,
    (void (*)(void)) extapp_scanKeyboard,
    (void (*)(void)) extapp_pushRect,
    (void (*)(void)) extapp_pushRectUniform, // 4
    (void (*)(void)) extapp_pullRect,
    (void (*)(void)) extapp_drawTextLarge,
    (void (*)(void)) extapp_drawTextSmall,
    (void (*)(void)) extapp_waitForVBlank, // 8
    (void (*)(void)) extapp_clipboardStore,
    (void (*)(void)) extapp_clipboardText,
    (void (*)(void)) extapp_fileListWithExtension,
    (void (*)(void)) extapp_fileExists, // 12
    (void (*)(void)) extapp_fileErase,
    (void (*)(void)) extapp_fileRead,
    (void (*)(void)) extapp_fileWrite,
    (void (*)(void)) extapp_lockAlpha, // 16
    (void (*)(void)) extapp_resetKeyboard,
    (void (*)(void)) extapp_getKey,
    (void (*)(void)) extapp_restorebackup,
    (void (*)(void)) extapp_erasesector, // 20
    (void (*)(void)) extapp_writememory,
    (void (*)(void)) extapp_inexammode,
    (void (*)(void)) extapp_isalphaactive, 
    (void (*)(void)) extapp_kbdstatus,  // 24
    (void (*)(void)) nullptr,
  };
  
  int main(int argc, char * argv[]) {
    const size_t * userland=extapp_userlandAddress();
    size_t beginapp=userland[7],endapp=userland[8];
    // check if there is enough room for data/bss linked at 0x24003300 or 0x20003300 with size=13K
    if ( (endapp & 0x3ffff) < 13*1024+0x33000){
      extapp_drawTextSmall("",100,100,65535,0,false);
      extapp_getKey(false,0);      
      return -1;
    }
    int heapsize=heapSize;
    if (endapp-beginapp <= heapsize+13*1024)
      heapsize=endapp-beginapp-13*1024;
    if (heapsize<64*1024){
      extapp_drawTextSmall("Not enough space for heap",0,100,65535,0,false);
      extapp_getKey(false,0);      
      return -2;
    }
    uint32_t ep = *reinterpret_cast<const uint32_t*>(ENTRYPOINT);
    char * heap=(char *)malloc(heapsize);
    while (heap==0){
      if (heapsize<64*1024){
        extapp_drawTextSmall("Not enough space for heap",0,100,65535,0,false);
        extapp_getKey(false,0);      
        return -3;
      }
      heapsize -= 1024;
      heap=(char *)malloc(heapsize);
    }
    extapp_pushRectUniform(0,0,320,240,0);
    extapp_pushRectUniform(0,0,320,18,0x7ff);
    extapp_drawTextSmall("Press any key to run KhiCAS (beta)",20,20,65535,0,false);
    extapp_drawTextSmall("Press HOME to leave",20,40,65535,0,false);
    extapp_drawTextSmall("Launcher parameters",0,60,65535,0,false);
    extapp_drawTextSmall("Begin app            ",0,80,65535,0,false);
    char tab[9]={0};
    print_hex(beginapp,tab);
    extapp_drawTextSmall(tab,100,80,65535,0,false);
    extapp_drawTextSmall("End app            ",0,100,65535,0,false);
    char tab1[9]={0};
    print_hex(endapp,tab1);
    extapp_drawTextSmall(tab1,100,100,65535,0,false);
    extapp_drawTextSmall("Appfree-13K        ",0,120,65535,0,false);
    char tab2[9]={0};
    print_hex(endapp-beginapp-13*1024,tab2);
    extapp_drawTextSmall(tab2,100,120,65535,0,false);
    extapp_drawTextSmall("Entrypoint            ",0,140,65535,0,false);
    char tab3[9]={0};
    print_hex(ep,tab3);
    extapp_drawTextSmall(tab3,100,140,65535,0,false);
    extapp_drawTextSmall("Heap            ",0,160,65535,0,false);
    char tab4[9]={0};
    print_hex((size_t)heap,tab4);
    extapp_drawTextSmall(tab4,100,160,65535,0,false);
#if 0 // beta testing
    extapp_getKey(false,0);
#endif
    if (ep >= 0x90400000 && ep < 0x90800000)
      return ((entrypoint)ep)(API_VERSION, apiPointers, heap, heapsize);
    else {
      extapp_drawTextSmall("Invalid entrypoint. Reinstall KhiCAS:",20,180,0xf800,0,false);
      extapp_drawTextSmall("xcas.univ-grenoble-alpes.fr/nw/nws.html",20,200,65535,0,false); 
      extapp_getKey(false,0);
    }
    return -2;
  }

}

